@@ -16,13 +16,11 @@
 #
 
 if [[ ${TRAVIS_OS_NAME} = "linux" ]]; then
-  sudo add-apt-repository ppa:jbboehr/build-deps -y # For re2c 0.16.
   sudo apt-get update -qq
-  sudo apt-get install -qq re2c gcc-4.8-multilib g++-4.8-multilib -y
+  sudo apt-get install -qq gcc-4.8-multilib g++-4.8-multilib -y
   sudo pip install flake8
 elif [[ ${TRAVIS_OS_NAME} = "osx" ]]; then
   brew update
-  brew install re2c
 else
   echo "unknown TRAVIS_OS_NAME: ${TRAVIS_OS_NAME}"
   exit 1
