@@ -42,6 +42,7 @@ class NameApplier : public ExprVisitor::DelegateNop {
   Result OnBrOnExnExpr(BrOnExnExpr*) override;
   Result OnBrTableExpr(BrTableExpr*) override;
   Result OnCallExpr(CallExpr*) override;
+  Result OnRefFuncExpr(RefFuncExpr*) override;
   Result OnCallIndirectExpr(CallIndirectExpr*) override;
   Result OnReturnCallExpr(ReturnCallExpr*) override;
   Result OnReturnCallIndirectExpr(ReturnCallIndirectExpr*) override;
@@ -62,6 +63,7 @@ class NameApplier : public ExprVisitor::DelegateNop {
   Result OnTableSetExpr(TableSetExpr*) override;
   Result OnTableGrowExpr(TableGrowExpr*) override;
   Result OnTableSizeExpr(TableSizeExpr*) override;
+  Result OnTableFillExpr(TableFillExpr*) override;
   Result BeginTryExpr(TryExpr*) override;
   Result EndTryExpr(TryExpr*) override;
   Result OnThrowExpr(ThrowExpr*) override;
@@ -86,6 +88,7 @@ class NameApplier : public ExprVisitor::DelegateNop {
   Result VisitExport(Index export_index, Export* export_);
   Result VisitElemSegment(Index elem_segment_index, ElemSegment* segment);
   Result VisitDataSegment(Index data_segment_index, DataSegment* segment);
+  Result VisitStart(Var* start_var);
 
   Module* module_ = nullptr;
   Func* current_func_ = nullptr;
@@ -258,7 +261,8 @@ Result NameApplier::OnElemDropExpr(ElemDropExpr* expr)  {
 }
 
 Result NameApplier::OnTableInitExpr(TableInitExpr* expr)  {
-  CHECK_RESULT(UseNameForElemSegmentVar(&expr->var));
+  CHECK_RESULT(UseNameForElemSegmentVar(&expr->segment_index));
+  CHECK_RESULT(UseNameForTableVar(&expr->table_index));
   return Result::Ok;
 }
 
@@ -282,6 +286,11 @@ Result NameApplier::OnTableSizeExpr(TableSizeExpr* expr)  {
   return Result::Ok;
 }
 
+Result NameApplier::OnTableFillExpr(TableFillExpr* expr)  {
+  CHECK_RESULT(UseNameForTableVar(&expr->var));
+  return Result::Ok;
+}
+
 Result NameApplier::OnBrExpr(BrExpr* expr) {
   string_view label = FindLabelByVar(&expr->var);
   UseNameForVar(label, &expr->var);
@@ -332,6 +341,11 @@ Result NameApplier::OnCallExpr(CallExpr* expr) {
   return Result::Ok;
 }
 
+Result NameApplier::OnRefFuncExpr(RefFuncExpr* expr) {
+  CHECK_RESULT(UseNameForFuncVar(&expr->var));
+  return Result::Ok;
+}
+
 Result NameApplier::OnCallIndirectExpr(CallIndirectExpr* expr) {
   if (expr->decl.has_func_type) {
     CHECK_RESULT(UseNameForFuncTypeVar(&expr->decl.type_var));
@@ -440,6 +454,11 @@ Result NameApplier::VisitDataSegment(Index data_segment_index,
   return Result::Ok;
 }
 
+Result NameApplier::VisitStart(Var* start_var) {
+  CHECK_RESULT(UseNameForFuncVar(start_var));
+  return Result::Ok;
+}
+
 Result NameApplier::VisitModule(Module* module) {
   module_ = module;
   for (size_t i = 0; i < module->funcs.size(); ++i)
@@ -454,6 +473,8 @@ Result NameApplier::VisitModule(Module* module) {
     CHECK_RESULT(VisitElemSegment(i, module->elem_segments[i]));
   for (size_t i = 0; i < module->data_segments.size(); ++i)
     CHECK_RESULT(VisitDataSegment(i, module->data_segments[i]));
+  for (size_t i = 0; i < module->starts.size(); ++i)
+    CHECK_RESULT(VisitStart(module->starts[i]));
   module_ = nullptr;
   return Result::Ok;
 }
