@@ -75,12 +75,6 @@ enum class NameSectionSubsection {
   Local = 2,
 };
 
-enum class SegmentFlags : uint8_t {
-  IndexZero = 0,
-  Passive = 1,
-  IndexOther = 2,
-};
-
 }  // namespace wabt
 
 #endif /* WABT_BINARY_H_ */
