@@ -27,7 +27,9 @@ class Stream;
 struct DecompileOptions {
 };
 
-Result Decompile(Stream&, const Module&, const DecompileOptions&);
+void RenameAll(Module&);
+
+std::string Decompile(const Module&, const DecompileOptions&);
 
 }  // namespace wabt
 
