@@ -24,6 +24,7 @@ namespace wabt {
 
 namespace interp {
 
+struct Export;
 struct DefinedModule;
 class Environment;
 
@@ -31,6 +32,7 @@ class Environment;
 
 struct ReadBinaryOptions;
 
+// Read and instantiate a module in the given environment.
 Result ReadBinaryInterp(interp::Environment* env,
                         const void* data,
                         size_t size,
@@ -38,6 +40,16 @@ Result ReadBinaryInterp(interp::Environment* env,
                         Errors*,
                         interp::DefinedModule** out_module);
 
+// Read and instantiate a module in the given environment. Similar to above but
+// using using a fixed set of exports to resolve the module exports.
+Result ReadBinaryInterp(interp::Environment* env,
+                        const void* data,
+                        size_t size,
+                        const ReadBinaryOptions& options,
+                        const std::vector<interp::Export*>& imports,
+                        Errors*,
+                        interp::DefinedModule** out_module);
+
 }  // namespace wabt
 
 #endif /* WABT_BINARY_READER_INTERP_H_ */
