@@ -43,6 +43,7 @@ void Environment::Disassemble(Stream* stream,
     assert(!opcode.IsInvalid());
     switch (opcode) {
       case Opcode::Select:
+      case Opcode::SelectT:
       case Opcode::V128BitSelect:
         stream->Writef("%s %%[-3], %%[-2], %%[-1]\n", opcode.GetName());
         break;
@@ -143,7 +144,17 @@ void Environment::Disassemble(Stream* stream,
       case Opcode::I64Load:
       case Opcode::F32Load:
       case Opcode::F64Load:
-      case Opcode::V128Load: {
+      case Opcode::V128Load:
+      case Opcode::I8X16LoadSplat:
+      case Opcode::I16X8LoadSplat:
+      case Opcode::I32X4LoadSplat:
+      case Opcode::I64X2LoadSplat:
+      case Opcode::I16X8Load8X8S:
+      case Opcode::I16X8Load8X8U:
+      case Opcode::I32X4Load16X4S:
+      case Opcode::I32X4Load16X4U:
+      case Opcode::I64X2Load32X2S:
+      case Opcode::I64X2Load32X2U: {
         const Index memory_index = ReadU32(&pc);
         stream->Writef("%s $%" PRIindex ":%%[-1]+$%u\n", opcode.GetName(),
                        memory_index, ReadU32(&pc));
@@ -395,6 +406,14 @@ void Environment::Disassemble(Stream* stream,
       case Opcode::F64X2Div:
       case Opcode::F32X4Mul:
       case Opcode::F64X2Mul:
+      case Opcode::V8X16Swizzle:
+      case Opcode::I8X16NarrowI16X8S:
+      case Opcode::I8X16NarrowI16X8U:
+      case Opcode::I16X8NarrowI32X4S:
+      case Opcode::I16X8NarrowI32X4U:
+      case Opcode::V128Andnot:
+      case Opcode::I8X16AvgrU:
+      case Opcode::I16X8AvgrU:
         stream->Writef("%s %%[-2], %%[-1]\n", opcode.GetName());
         break;
 
@@ -491,6 +510,14 @@ void Environment::Disassemble(Stream* stream,
       case Opcode::I32X4TruncSatF32X4U:
       case Opcode::I64X2TruncSatF64X2S:
       case Opcode::I64X2TruncSatF64X2U:
+      case Opcode::I16X8WidenLowI8X16S:
+      case Opcode::I16X8WidenHighI8X16S:
+      case Opcode::I16X8WidenLowI8X16U:
+      case Opcode::I16X8WidenHighI8X16U:
+      case Opcode::I32X4WidenLowI16X8S:
+      case Opcode::I32X4WidenHighI16X8S:
+      case Opcode::I32X4WidenLowI16X8U:
+      case Opcode::I32X4WidenHighI16X8U:
         stream->Writef("%s %%[-1]\n", opcode.GetName());
         break;
 
@@ -581,13 +608,32 @@ void Environment::Disassemble(Stream* stream,
       }
 
       case Opcode::TableGet:
+        stream->Writef("%s %u %%[-1]\n", opcode.GetName(), ReadU32(&pc));
+        break;
+
       case Opcode::TableSet:
+        stream->Writef("%s $%u %%[-1] %%[-2]\n", opcode.GetName(), ReadU32(&pc));
+        break;
+
       case Opcode::TableGrow:
       case Opcode::TableSize:
+        break;
+
+      case Opcode::TableFill:
+        stream->Writef("%s %u %%[-3], %%[-2], %%[-1]\n", opcode.GetName(),
+                       ReadU32(&pc));
+        break;
+
       case Opcode::RefNull:
-      case Opcode::RefIsNull:
+        stream->Writef("%s\n", opcode.GetName());
+        break;
+
       case Opcode::RefFunc:
-        WABT_UNREACHABLE;
+        stream->Writef("%s $%u\n", opcode.GetName(), ReadU32(&pc));
+        break;
+
+      case Opcode::RefIsNull:
+        stream->Writef("%s %%[-1]\n", opcode.GetName());
         break;
 
       case Opcode::MemoryInit:
@@ -606,12 +652,16 @@ void Environment::Disassemble(Stream* stream,
         break;
 
       case Opcode::MemoryCopy:
-      case Opcode::TableCopy:
       case Opcode::MemoryFill:
         stream->Writef("%s $%u, %%[-3], %%[-2], %%[-1]\n", opcode.GetName(),
                        ReadU32(&pc));
         break;
 
+      case Opcode::TableCopy:
+        stream->Writef("%s $%u, $%u, %%[-3], %%[-2], %%[-1]\n",
+                       opcode.GetName(), ReadU32(&pc), ReadU32(&pc));
+        break;
+
       // The following opcodes are either never generated or should never be
       // executed.
       case Opcode::Block:
@@ -625,6 +675,7 @@ void Environment::Disassemble(Stream* stream,
       case Opcode::Rethrow:
       case Opcode::Throw:
       case Opcode::Try:
+        fprintf(stderr, "unknown opcode: %#x\n", static_cast<uint32_t>(opcode));
         WABT_UNREACHABLE;
         break;
     }
