@@ -34,6 +34,7 @@ void Thread::Trace(Stream* stream) {
   assert(!opcode.IsInvalid());
   switch (opcode) {
     case Opcode::Select:
+    case Opcode::SelectT:
       // TODO(binji): We don't know the type here so we can't display the value
       // to the user. This used to display the full 64-bit value, but that
       // will potentially display garbage if the value is 32-bit.
@@ -141,7 +142,13 @@ void Thread::Trace(Stream* stream) {
     case Opcode::I8X16LoadSplat:
     case Opcode::I16X8LoadSplat:
     case Opcode::I32X4LoadSplat:
-    case Opcode::I64X2LoadSplat: {
+    case Opcode::I64X2LoadSplat:
+    case Opcode::I16X8Load8X8S:
+    case Opcode::I16X8Load8X8U:
+    case Opcode::I32X4Load16X4S:
+    case Opcode::I32X4Load16X4U:
+    case Opcode::I64X2Load32X2S:
+    case Opcode::I64X2Load32X2U: {
       const Index memory_index = ReadU32(&pc);
       stream->Writef("%s $%" PRIindex ":%u+$%u\n", opcode.GetName(),
                      memory_index, Top().i32, ReadU32At(pc));
@@ -268,8 +275,8 @@ void Thread::Trace(Stream* stream) {
       const Index memory_index = ReadU32(&pc);
       stream->Writef("%s $%" PRIindex ":%u+$%u, $0x%08x 0x%08x 0x%08x 0x%08x\n",
                      opcode.GetName(), memory_index, Pick(2).i32, ReadU32At(pc),
-                     Pick(1).v128_bits.v[0], Pick(1).v128_bits.v[1],
-                     Pick(1).v128_bits.v[2], Pick(1).v128_bits.v[3]);
+                     Pick(1).vec128.v[0], Pick(1).vec128.v[1],
+                     Pick(1).vec128.v[2], Pick(1).vec128.v[3]);
       break;
     }
 
@@ -509,10 +516,18 @@ void Thread::Trace(Stream* stream) {
     case Opcode::I32X4TruncSatF32X4S:
     case Opcode::I32X4TruncSatF32X4U:
     case Opcode::I64X2TruncSatF64X2S:
-    case Opcode::I64X2TruncSatF64X2U: {
+    case Opcode::I64X2TruncSatF64X2U:
+    case Opcode::I16X8WidenLowI8X16S:
+    case Opcode::I16X8WidenHighI8X16S:
+    case Opcode::I16X8WidenLowI8X16U:
+    case Opcode::I16X8WidenHighI8X16U:
+    case Opcode::I32X4WidenLowI16X8S:
+    case Opcode::I32X4WidenHighI16X8S:
+    case Opcode::I32X4WidenLowI16X8U:
+    case Opcode::I32X4WidenHighI16X8U: {
       stream->Writef("%s $0x%08x 0x%08x 0x%08x 0x%08x\n", opcode.GetName(),
-                     Top().v128_bits.v[0], Top().v128_bits.v[1],
-                     Top().v128_bits.v[2], Top().v128_bits.v[3]);
+                     Top().vec128.v[0], Top().vec128.v[1],
+                     Top().vec128.v[2], Top().vec128.v[3]);
       break;
     }
 
@@ -520,12 +535,12 @@ void Thread::Trace(Stream* stream) {
       stream->Writef(
           "%s $0x%08x %08x %08x %08x $0x%08x %08x %08x %08x $0x%08x %08x %08x "
           "%08x\n",
-          opcode.GetName(), Pick(3).v128_bits.v[0], Pick(3).v128_bits.v[1],
-          Pick(3).v128_bits.v[2], Pick(3).v128_bits.v[3],
-          Pick(2).v128_bits.v[0], Pick(2).v128_bits.v[1],
-          Pick(2).v128_bits.v[2], Pick(2).v128_bits.v[3],
-          Pick(1).v128_bits.v[0], Pick(1).v128_bits.v[1],
-          Pick(1).v128_bits.v[2], Pick(1).v128_bits.v[3]);
+          opcode.GetName(), Pick(3).vec128.v[0], Pick(3).vec128.v[1],
+          Pick(3).vec128.v[2], Pick(3).vec128.v[3],
+          Pick(2).vec128.v[0], Pick(2).vec128.v[1],
+          Pick(2).vec128.v[2], Pick(2).vec128.v[3],
+          Pick(1).vec128.v[0], Pick(1).vec128.v[1],
+          Pick(1).vec128.v[2], Pick(1).vec128.v[3]);
       break;
 
     case Opcode::I8X16ExtractLaneS:
@@ -537,9 +552,9 @@ void Thread::Trace(Stream* stream) {
     case Opcode::F32X4ExtractLane:
     case Opcode::F64X2ExtractLane: {
       stream->Writef("%s : LaneIdx %d From $0x%08x 0x%08x 0x%08x 0x%08x\n",
-                     opcode.GetName(), ReadU8At(pc), Top().v128_bits.v[0],
-                     Top().v128_bits.v[1], Top().v128_bits.v[2],
-                     Top().v128_bits.v[3]);
+                     opcode.GetName(), ReadU8At(pc), Top().vec128.v[0],
+                     Top().vec128.v[1], Top().vec128.v[2],
+                     Top().vec128.v[3]);
       break;
     }
 
@@ -548,25 +563,25 @@ void Thread::Trace(Stream* stream) {
     case Opcode::I32X4ReplaceLane: {
       stream->Writef(
           "%s : Set %u to LaneIdx %d In $0x%08x 0x%08x 0x%08x 0x%08x\n",
-          opcode.GetName(), Pick(1).i32, ReadU8At(pc), Pick(2).v128_bits.v[0],
-          Pick(2).v128_bits.v[1], Pick(2).v128_bits.v[2],
-          Pick(2).v128_bits.v[3]);
+          opcode.GetName(), Pick(1).i32, ReadU8At(pc), Pick(2).vec128.v[0],
+          Pick(2).vec128.v[1], Pick(2).vec128.v[2],
+          Pick(2).vec128.v[3]);
       break;
     }
     case Opcode::I64X2ReplaceLane: {
       stream->Writef("%s : Set %" PRIu64
                      " to LaneIdx %d In $0x%08x 0x%08x 0x%08x 0x%08x\n",
                      opcode.GetName(), Pick(1).i64, ReadU8At(pc),
-                     Pick(2).v128_bits.v[0], Pick(2).v128_bits.v[1],
-                     Pick(2).v128_bits.v[2], Pick(2).v128_bits.v[3]);
+                     Pick(2).vec128.v[0], Pick(2).vec128.v[1],
+                     Pick(2).vec128.v[2], Pick(2).vec128.v[3]);
       break;
     }
     case Opcode::F32X4ReplaceLane: {
       stream->Writef(
           "%s : Set %g to LaneIdx %d In $0x%08x 0x%08x 0x%08x 0x%08x\n",
           opcode.GetName(), Bitcast<float>(Pick(1).f32_bits), ReadU8At(pc),
-          Pick(2).v128_bits.v[0], Pick(2).v128_bits.v[1],
-          Pick(2).v128_bits.v[2], Pick(2).v128_bits.v[3]);
+          Pick(2).vec128.v[0], Pick(2).vec128.v[1],
+          Pick(2).vec128.v[2], Pick(2).vec128.v[3]);
 
       break;
     }
@@ -574,8 +589,8 @@ void Thread::Trace(Stream* stream) {
       stream->Writef(
           "%s : Set %g to LaneIdx %d In $0x%08x 0x%08x 0x%08x 0x%08x\n",
           opcode.GetName(), Bitcast<double>(Pick(1).f64_bits), ReadU8At(pc),
-          Pick(2).v128_bits.v[0], Pick(2).v128_bits.v[1],
-          Pick(2).v128_bits.v[2], Pick(2).v128_bits.v[3]);
+          Pick(2).vec128.v[0], Pick(2).vec128.v[1],
+          Pick(2).vec128.v[2], Pick(2).vec128.v[3]);
       break;
     }
 
@@ -583,10 +598,10 @@ void Thread::Trace(Stream* stream) {
       stream->Writef(
           "%s $0x%08x %08x %08x %08x $0x%08x %08x %08x %08x : with lane imm: "
           "$0x%08x %08x %08x %08x\n",
-          opcode.GetName(), Pick(2).v128_bits.v[0], Pick(2).v128_bits.v[1],
-          Pick(2).v128_bits.v[2], Pick(2).v128_bits.v[3],
-          Pick(1).v128_bits.v[0], Pick(1).v128_bits.v[1],
-          Pick(1).v128_bits.v[2], Pick(1).v128_bits.v[3], ReadU32At(pc),
+          opcode.GetName(), Pick(2).vec128.v[0], Pick(2).vec128.v[1],
+          Pick(2).vec128.v[2], Pick(2).vec128.v[3],
+          Pick(1).vec128.v[0], Pick(1).vec128.v[1],
+          Pick(1).vec128.v[2], Pick(1).vec128.v[3], ReadU32At(pc),
           ReadU32At(pc + 4), ReadU32At(pc + 8), ReadU32At(pc + 12));
       break;
 
@@ -666,13 +681,20 @@ void Thread::Trace(Stream* stream) {
     case Opcode::F64X2Div:
     case Opcode::F32X4Mul:
     case Opcode::F64X2Mul:
-    case Opcode::V8X16Swizzle: {
+    case Opcode::V8X16Swizzle:
+    case Opcode::I8X16NarrowI16X8S:
+    case Opcode::I8X16NarrowI16X8U:
+    case Opcode::I16X8NarrowI32X4S:
+    case Opcode::I16X8NarrowI32X4U:
+    case Opcode::V128Andnot:
+    case Opcode::I8X16AvgrU:
+    case Opcode::I16X8AvgrU: {
       stream->Writef("%s $0x%08x %08x %08x %08x  $0x%08x %08x %08x %08x\n",
-                     opcode.GetName(), Pick(2).v128_bits.v[0],
-                     Pick(2).v128_bits.v[1], Pick(2).v128_bits.v[2],
-                     Pick(2).v128_bits.v[3], Pick(1).v128_bits.v[0],
-                     Pick(1).v128_bits.v[1], Pick(1).v128_bits.v[2],
-                     Pick(1).v128_bits.v[3]);
+                     opcode.GetName(), Pick(2).vec128.v[0],
+                     Pick(2).vec128.v[1], Pick(2).vec128.v[2],
+                     Pick(2).vec128.v[3], Pick(1).vec128.v[0],
+                     Pick(1).vec128.v[1], Pick(1).vec128.v[2],
+                     Pick(1).vec128.v[3]);
       break;
     }
 
@@ -689,20 +711,24 @@ void Thread::Trace(Stream* stream) {
     case Opcode::I64X2ShrS:
     case Opcode::I64X2ShrU: {
       stream->Writef("%s $0x%08x %08x %08x %08x  $0x%08x\n", opcode.GetName(),
-                     Pick(2).v128_bits.v[0], Pick(2).v128_bits.v[1],
-                     Pick(2).v128_bits.v[2], Pick(2).v128_bits.v[3],
+                     Pick(2).vec128.v[0], Pick(2).vec128.v[1],
+                     Pick(2).vec128.v[2], Pick(2).vec128.v[3],
                      Pick(1).i32);
       break;
     }
 
-    case Opcode::TableGet:
-    case Opcode::TableSet:
-    case Opcode::TableGrow:
-    case Opcode::TableSize:
     case Opcode::RefNull:
+      stream->Writef("%s\n", opcode.GetName());
+      break;
+
     case Opcode::RefIsNull:
+      stream->Writef("%s %s:%08x\n", opcode.GetName(),
+                     RefTypeToString(Pick(1).ref.kind).c_str(),
+                     Pick(1).ref.index);
+      break;
+
     case Opcode::RefFunc:
-      WABT_UNREACHABLE;
+      stream->Writef("%s $%u\n", opcode.GetName(), ReadU32At(pc));
       break;
 
     case Opcode::MemoryInit:
@@ -714,6 +740,11 @@ void Thread::Trace(Stream* stream) {
       break;
     }
 
+    case Opcode::TableGet:
+    case Opcode::TableSet:
+    case Opcode::TableSize:
+    case Opcode::TableFill:
+    case Opcode::TableGrow:
     case Opcode::DataDrop:
     case Opcode::ElemDrop:
     case Opcode::MemoryCopy:
