@@ -33,6 +33,8 @@
 
 namespace wabt {
 
+struct Module;
+
 enum class VarType {
   Index,
   Name,
@@ -96,13 +98,16 @@ struct Const {
 
   Location loc;
   Type type;
+
+  bool is_expected_nan = false;
   union {
     uint32_t u32;
     uint64_t u64;
     uint32_t f32_bits;
     uint64_t f64_bits;
     uintptr_t ref_bits;
-    v128 v128_bits;
+    v128 vec128;
+    ExpectedNan expected;
   };
 
  private:
@@ -193,6 +198,7 @@ enum class ExprType {
   MemorySize,
   Nop,
   RefIsNull,
+  RefFunc,
   RefNull,
   Rethrow,
   Return,
@@ -210,6 +216,7 @@ enum class ExprType {
   TableGrow,
   TableSize,
   TableSet,
+  TableFill,
   Ternary,
   Throw,
   Try,
@@ -269,11 +276,9 @@ typedef ExprMixin<ExprType::MemoryGrow> MemoryGrowExpr;
 typedef ExprMixin<ExprType::MemorySize> MemorySizeExpr;
 typedef ExprMixin<ExprType::MemoryCopy> MemoryCopyExpr;
 typedef ExprMixin<ExprType::MemoryFill> MemoryFillExpr;
-typedef ExprMixin<ExprType::TableCopy> TableCopyExpr;
 typedef ExprMixin<ExprType::Nop> NopExpr;
 typedef ExprMixin<ExprType::Rethrow> RethrowExpr;
 typedef ExprMixin<ExprType::Return> ReturnExpr;
-typedef ExprMixin<ExprType::Select> SelectExpr;
 typedef ExprMixin<ExprType::Unreachable> UnreachableExpr;
 typedef ExprMixin<ExprType::RefNull> RefNullExpr;
 typedef ExprMixin<ExprType::RefIsNull> RefIsNullExpr;
@@ -323,6 +328,7 @@ class VarExpr : public ExprMixin<TypeEnum> {
 typedef VarExpr<ExprType::Br> BrExpr;
 typedef VarExpr<ExprType::BrIf> BrIfExpr;
 typedef VarExpr<ExprType::Call> CallExpr;
+typedef VarExpr<ExprType::RefFunc> RefFuncExpr;
 typedef VarExpr<ExprType::GlobalGet> GlobalGetExpr;
 typedef VarExpr<ExprType::GlobalSet> GlobalSetExpr;
 typedef VarExpr<ExprType::LocalGet> LocalGetExpr;
@@ -333,12 +339,43 @@ typedef VarExpr<ExprType::Throw> ThrowExpr;
 
 typedef VarExpr<ExprType::MemoryInit> MemoryInitExpr;
 typedef VarExpr<ExprType::DataDrop> DataDropExpr;
-typedef VarExpr<ExprType::TableInit> TableInitExpr;
 typedef VarExpr<ExprType::ElemDrop> ElemDropExpr;
 typedef VarExpr<ExprType::TableGet> TableGetExpr;
 typedef VarExpr<ExprType::TableSet> TableSetExpr;
 typedef VarExpr<ExprType::TableGrow> TableGrowExpr;
 typedef VarExpr<ExprType::TableSize> TableSizeExpr;
+typedef VarExpr<ExprType::TableFill> TableFillExpr;
+
+class SelectExpr : public ExprMixin<ExprType::Select> {
+ public:
+  SelectExpr(TypeVector type, const Location& loc = Location())
+      : ExprMixin<ExprType::Select>(loc), result_type(type) {}
+  TypeVector result_type;
+};
+
+class TableInitExpr : public ExprMixin<ExprType::TableInit> {
+ public:
+  TableInitExpr(const Var& segment_index,
+                const Var& table_index,
+                const Location& loc = Location())
+      : ExprMixin<ExprType::TableInit>(loc),
+        segment_index(segment_index),
+        table_index(table_index) {}
+
+  Var segment_index;
+  Var table_index;
+};
+
+class TableCopyExpr : public ExprMixin<ExprType::TableCopy> {
+ public:
+  TableCopyExpr(const Var& dst,
+                const Var& src,
+                const Location& loc = Location())
+      : ExprMixin<ExprType::TableCopy>(loc), dst_table(dst), src_table(src) {}
+
+  Var dst_table;
+  Var src_table;
+};
 
 class CallIndirectExpr : public ExprMixin<ExprType::CallIndirect> {
  public:
@@ -568,11 +605,12 @@ typedef std::vector<ElemExpr> ElemExprVector;
 
 struct ElemSegment {
   explicit ElemSegment(string_view name) : name(name.to_string()) {}
+  uint8_t GetFlags(const Module*) const;
 
+  SegmentKind kind = SegmentKind::Active;
   std::string name;
   Var table_var;
-  bool passive = false;
-  Type elem_type;  // If passive == false, this is always Type::Funcref.
+  Type elem_type;
   ExprList offset;
   ElemExprVector elem_exprs;
 };
@@ -586,10 +624,11 @@ struct Memory {
 
 struct DataSegment {
   explicit DataSegment(string_view name) : name(name.to_string()) {}
+  uint8_t GetFlags(const Module*) const;
 
+  SegmentKind kind = SegmentKind::Active;
   std::string name;
   Var memory_var;
-  bool passive = false;
   ExprList offset;
   std::vector<uint8_t> data;
 };
@@ -1010,8 +1049,6 @@ enum class CommandType {
   AssertUnlinkable,
   AssertUninstantiable,
   AssertReturn,
-  AssertReturnCanonicalNan,
-  AssertReturnArithmeticNan,
   AssertTrap,
   AssertExhaustion,
 
@@ -1051,10 +1088,6 @@ class ActionCommandBase : public CommandMixin<TypeEnum> {
 };
 
 typedef ActionCommandBase<CommandType::Action> ActionCommand;
-typedef ActionCommandBase<CommandType::AssertReturnCanonicalNan>
-    AssertReturnCanonicalNanCommand;
-typedef ActionCommandBase<CommandType::AssertReturnArithmeticNan>
-    AssertReturnArithmeticNanCommand;
 
 class RegisterCommand : public CommandMixin<CommandType::Register> {
  public:
