@@ -23,18 +23,18 @@ namespace wabt {
 // static
 Opcode::Info Opcode::infos_[] = {
 #define WABT_OPCODE(rtype, type1, type2, type3, mem_size, prefix, code, Name, \
-                    text)                                                     \
-  {text,        Type::rtype, Type::type1,                                     \
+                    text, decomp)                                             \
+  {text, decomp, Type::rtype, Type::type1,                                    \
    Type::type2, Type::type3, mem_size,                                        \
    prefix,      code,        PrefixCode(prefix, code)},
 #include "src/opcode.def"
 #undef WABT_OPCODE
 
-    {"<invalid>", Type::Void, Type::Void, Type::Void, Type::Void, 0, 0, 0, 0},
+  {"<invalid>", "", Type::Void, Type::Void, Type::Void, Type::Void, 0, 0, 0, 0},
 };
 
 #define WABT_OPCODE(rtype, type1, type2, type3, mem_size, prefix, code, Name, \
-                    text)                                                     \
+                    text, decomp)                                                     \
   /* static */ Opcode Opcode::Name##_Opcode(Opcode::Name);
 #include "src/opcode.def"
 #undef WABT_OPCODE
@@ -183,7 +183,6 @@ bool Opcode::IsEnabled(const Features& features) const {
     case Opcode::I64X2ReplaceLane:
     case Opcode::F32X4ReplaceLane:
     case Opcode::F64X2ReplaceLane:
-    case Opcode::V8X16Shuffle:
     case Opcode::I8X16Add:
     case Opcode::I16X8Add:
     case Opcode::I32X4Add:
@@ -300,6 +299,12 @@ bool Opcode::IsEnabled(const Features& features) const {
     case Opcode::I32X4TruncSatF32X4U:
     case Opcode::I64X2TruncSatF64X2S:
     case Opcode::I64X2TruncSatF64X2U:
+    case Opcode::V8X16Swizzle:
+    case Opcode::V8X16Shuffle:
+    case Opcode::I8X16LoadSplat:
+    case Opcode::I16X8LoadSplat:
+    case Opcode::I32X4LoadSplat:
+    case Opcode::I64X2LoadSplat:
       return features.simd_enabled();
 
     case Opcode::MemoryInit:
