@@ -46,6 +46,7 @@ static Thread::Options s_thread_options;
 static Stream* s_trace_stream;
 static bool s_run_all_exports;
 static bool s_host_print;
+static bool s_dummy_import_func;
 static Features s_features;
 
 static std::unique_ptr<FileStream> s_log_stream;
@@ -82,7 +83,6 @@ static void ParseOptions(int argc, char** argv) {
     s_verbose++;
     s_log_stream = FileStream::CreateStdout();
   });
-  parser.AddHelpOption();
   s_features.AddOptions(&parser);
   parser.AddOption('V', "value-stack-size", "SIZE",
                    "Size in elements of the value stack",
@@ -106,6 +106,11 @@ static void ParseOptions(int argc, char** argv) {
                    "Include an importable function named \"host.print\" for "
                    "printing to stdout",
                    []() { s_host_print = true; });
+  parser.AddOption(
+      "dummy-import-func",
+      "Provide a dummy implementation of all imported functions. The function "
+      "will log the call and return an appropriate zero value.",
+      []() { s_dummy_import_func = true; });
 
   parser.AddArgument("filename", OptionParser::ArgumentCount::One,
                      [](const char* argument) { s_infile = argument; });
@@ -163,13 +168,13 @@ static interp::Result PrintCallback(const HostFunc* func,
                                     TypedValues& results) {
   printf("called host ");
   WriteCall(s_stdout_stream.get(), func->module_name, func->field_name, args,
-            results, interp::Result::Ok);
-  return interp::Result::Ok;
+            results, interp::ResultType::Ok);
+  return interp::ResultType::Ok;
 }
 
 static void InitEnvironment(Environment* env) {
   if (s_host_print) {
-    HostModule* host_module = env->AppendHostModule("host");
+    auto* host_module = env->AppendHostModule("host");
     host_module->on_unknown_func_export =
         [](Environment* env, HostModule* host_module, string_view name,
            Index sig_index) -> Index {
@@ -177,16 +182,28 @@ static void InitEnvironment(Environment* env) {
         return kInvalidIndex;
       }
 
-      std::pair<HostFunc*, Index> pair =
-          host_module->AppendFuncExport(name, sig_index, PrintCallback);
-      return pair.second;
+      return host_module->AppendFuncExport(name, sig_index, PrintCallback)
+          .second;
+    };
+  }
+
+  if (s_dummy_import_func) {
+    env->on_unknown_module = [](Environment* env, string_view name) {
+      auto* host_module = env->AppendHostModule(name);
+      host_module->on_unknown_func_export =
+          [](Environment* env, HostModule* host_module, string_view name,
+             Index sig_index) -> Index {
+        return host_module->AppendFuncExport(name, sig_index, PrintCallback)
+            .second;
+      };
+      return true;
     };
   }
 }
 
 static wabt::Result ReadAndRunModule(const char* module_filename) {
   wabt::Result result;
-  Environment env;
+  Environment env(s_features);
   InitEnvironment(&env);
 
   Errors errors;
@@ -195,14 +212,15 @@ static wabt::Result ReadAndRunModule(const char* module_filename) {
   FormatErrorsToFile(errors, Location::Type::Binary);
   if (Succeeded(result)) {
     Executor executor(&env, s_trace_stream, s_thread_options);
-    ExecResult exec_result = executor.RunStartFunction(module);
-    if (exec_result.result == interp::Result::Ok) {
+    ExecResult exec_result = executor.Initialize(module);
+    if (exec_result.ok()) {
       if (s_run_all_exports) {
         RunAllExports(module, &executor, RunVerbosity::Verbose);
       }
     } else {
-      WriteResult(s_stdout_stream.get(), "error running start function",
+      WriteResult(s_stdout_stream.get(), "error initialiazing module",
                   exec_result.result);
+      return wabt::Result::Error;
     }
   }
   return result;
