@@ -51,9 +51,6 @@ TypeChecker::Label::Label(LabelType label_type,
       type_stack_limit(limit),
       unreachable(false) {}
 
-TypeChecker::TypeChecker(const ErrorCallback& error_callback)
-    : error_callback_(error_callback) {}
-
 void TypeChecker::PrintError(const char* fmt, ...) {
   if (error_callback_) {
     WABT_SNPRINTF_ALLOCA(buffer, length, fmt);
@@ -168,17 +165,32 @@ Result TypeChecker::CheckTypeStackEnd(const char* desc) {
   return result;
 }
 
+static bool IsSubtype(Type sub, Type super) {
+  if (super == sub) {
+    return true;
+  }
+  if (IsRefType(super) != IsRefType(sub)) {
+    return false;
+  }
+  if (super == Type::Anyref) {
+    return IsRefType(sub);
+  }
+  if (IsNullableRefType(super)) {
+    return sub == Type::Nullref;
+  }
+  return false;
+}
+
 Result TypeChecker::CheckType(Type actual, Type expected) {
-  if (expected == actual || expected == Type::Any || actual == Type::Any) {
+  if (expected == Type::Any || actual == Type::Any) {
     return Result::Ok;
   }
-  if (expected == Type::Anyref && (actual == Type::Funcref || actual == Type::Nullref)) {
-    return Result::Ok;
+
+  if (!IsSubtype(actual, expected)) {
+    return Result::Error;
   }
-  if ((expected == Type::Funcref || expected == Type::Anyref) && actual == Type::Nullref) {
-    return Result::Ok;
-  }
-  return Result::Error;
+
+  return Result::Ok;
 }
 
 Result TypeChecker::CheckTypes(const TypeVector& actual,
@@ -421,12 +433,23 @@ Result TypeChecker::OnBrTableTarget(Index depth) {
   // signatures.
   if (br_table_sig_ == nullptr) {
     br_table_sig_ = &label_sig;
-  }
-  if (*br_table_sig_ != label_sig) {
-    result |= Result::Error;
-    PrintError("br_table labels have inconsistent types: expected %s, got %s",
-               TypesToString(*br_table_sig_).c_str(),
-               TypesToString(label_sig).c_str());
+  } else {
+    if (features_.reference_types_enabled()) {
+      if (br_table_sig_->size() != label_sig.size()) {
+        result |= Result::Error;
+        PrintError("br_table labels have inconsistent arity: expected %" PRIzd
+                   " got %" PRIzd,
+                   br_table_sig_->size(), label_sig.size());
+      }
+    } else {
+      if (*br_table_sig_ != label_sig) {
+        result |= Result::Error;
+        PrintError(
+            "br_table labels have inconsistent types: expected %s, got %s",
+            TypesToString(*br_table_sig_).c_str(),
+            TypesToString(label_sig).c_str());
+      }
+    }
   }
 
   return result;
@@ -632,32 +655,40 @@ Result TypeChecker::OnElemDrop(uint32_t segment) {
   return Result::Ok;
 }
 
-Result TypeChecker::OnTableInit(uint32_t segment) {
+Result TypeChecker::OnTableInit(uint32_t table, uint32_t segment) {
   return CheckOpcode3(Opcode::TableInit);
 }
 
-Result TypeChecker::OnTableGet(Index segment) {
+Result TypeChecker::OnTableGet(Type elem_type) {
   Result result = PopAndCheck1Type(Type::I32, "table.get");
-  PushType(Type::Nullref); // TODO: should be the table's type
+  PushType(elem_type);
   return result;
 }
 
-Result TypeChecker::OnTableSet(Index segment) {
-  // TODO: anyref here should be the table's type
-  return PopAndCheck2Types(Type::I32, Type::Anyref, "table.set");
+Result TypeChecker::OnTableSet(Type elem_type) {
+  return PopAndCheck2Types(Type::I32, elem_type, "table.set");
 }
 
-Result TypeChecker::OnTableGrow(Index segment) {
-  Result result = PopAndCheck2Types(Type::Anyref, Type::I32, "table.grow");
+Result TypeChecker::OnTableGrow(Type elem_type) {
+  Result result = PopAndCheck2Types(elem_type, Type::I32, "table.grow");
   PushType(Type::I32);
   return result;
 }
 
-Result TypeChecker::OnTableSize(Index segment) {
+Result TypeChecker::OnTableSize() {
   PushType(Type::I32);
   return Result::Ok;
 }
 
+Result TypeChecker::OnTableFill(Type elem_type) {
+  return PopAndCheck3Types(Type::I32, elem_type, Type::I32, "table.fill");
+}
+
+Result TypeChecker::OnRefFuncExpr(Index) {
+  PushType(Type::Funcref);
+  return Result::Ok;
+}
+
 Result TypeChecker::OnRefNullExpr() {
   PushType(Type::Nullref);
   return Result::Ok;
@@ -691,15 +722,28 @@ Result TypeChecker::OnReturn() {
   return result;
 }
 
-Result TypeChecker::OnSelect() {
+Result TypeChecker::OnSelect(Type expected) {
   Result result = Result::Ok;
-  Type type = Type::Any;
+  Type type1 = Type::Any;
+  Type type2 = Type::Any;
+  Type result_type = Type::Any;
   result |= PeekAndCheckType(0, Type::I32);
-  result |= PeekType(1, &type);
-  result |= PeekAndCheckType(2, type);
-  PrintStackIfFailed(result, "select", Type::I32, type, type);
+  result |= PeekType(1, &type1);
+  result |= PeekType(2, &type2);
+  if (expected == Type::Any) {
+    if (IsRefType(type1) || IsRefType(type2)) {
+      result = Result::Error;
+    } else {
+      result |= CheckType(type1, type2);
+      result_type = type1;
+    }
+  } else {
+    result |= CheckType(type1, expected);
+    result |= CheckType(type2, expected);
+  }
+  PrintStackIfFailed(result, "select", result_type, result_type, Type::I32);
   result |= DropTypes(3);
-  PushType(type);
+  PushType(result_type);
   return result;
 }
 
