@@ -21,6 +21,7 @@
 #include <vector>
 
 #include "src/common.h"
+#include "src/feature.h"
 #include "src/opcode.h"
 
 namespace wabt {
@@ -46,8 +47,7 @@ class TypeChecker {
     bool unreachable;
   };
 
-  TypeChecker() = default;
-  explicit TypeChecker(const ErrorCallback&);
+  explicit TypeChecker(const Features& features) : features_(features) {}
 
   void set_error_callback(const ErrorCallback& error_callback) {
     error_callback_ = error_callback;
@@ -101,16 +101,18 @@ class TypeChecker {
   Result OnMemorySize();
   Result OnTableCopy();
   Result OnElemDrop(Index);
-  Result OnTableInit(Index);
-  Result OnTableGet(Index);
-  Result OnTableSet(Index);
-  Result OnTableGrow(Index);
-  Result OnTableSize(Index);
+  Result OnTableInit(Index, Index);
+  Result OnTableGet(Type elem_type);
+  Result OnTableSet(Type elem_type);
+  Result OnTableGrow(Type elem_type);
+  Result OnTableSize();
+  Result OnTableFill(Type elem_type);
+  Result OnRefFuncExpr(Index func_index);
   Result OnRefNullExpr();
   Result OnRefIsNullExpr();
   Result OnRethrow();
   Result OnReturn();
-  Result OnSelect();
+  Result OnSelect(Type expected);
   Result OnSimdLaneOp(Opcode, uint64_t);
   Result OnSimdShuffleOp(Opcode, v128);
   Result OnStore(Opcode);
@@ -175,6 +177,7 @@ class TypeChecker {
   // Cache the expected br_table signature. It will be initialized to `nullptr`
   // to represent "any".
   TypeVector* br_table_sig_ = nullptr;
+  Features features_;
 };
 
 }  // namespace wabt
