@@ -80,6 +80,8 @@ class Validator : public ExprVisitor::Delegate {
   Result OnTableSetExpr(TableSetExpr*) override;
   Result OnTableGrowExpr(TableGrowExpr*) override;
   Result OnTableSizeExpr(TableSizeExpr*) override;
+  Result OnTableFillExpr(TableFillExpr*) override;
+  Result OnRefFuncExpr(RefFuncExpr*) override;
   Result OnRefNullExpr(RefNullExpr*) override;
   Result OnRefIsNullExpr(RefIsNullExpr*) override;
   Result OnNopExpr(NopExpr*) override;
@@ -160,17 +162,10 @@ class Validator : public ExprVisitor::Delegate {
                   const TypeVector& expected,
                   const char* desc,
                   const char* index_kind);
-  void CheckConstTypes(const Location* loc,
-                       const TypeVector& actual,
-                       const ConstVector& expected,
-                       const char* desc);
-  void CheckConstType(const Location* loc,
-                      Type actual,
-                      const ConstVector& expected,
-                      const char* desc);
-  void CheckAssertReturnNanType(const Location* loc,
-                                Type actual,
-                                const char* desc);
+  void CheckResultTypes(const Location* loc,
+                        const TypeVector& actual,
+                        const TypeVector& expected,
+                        const char* desc);
   void CheckExprList(const Location* loc, const ExprList& exprs);
   bool CheckHasMemory(const Location* loc, Opcode opcode);
   bool CheckHasTable(const Location* loc, Opcode opcode, Index index = 0);
@@ -205,7 +200,6 @@ class Validator : public ExprVisitor::Delegate {
   const TypeVector* CheckInvoke(const InvokeAction* action);
   Result CheckGet(const GetAction* action, Type* out_type);
   ActionResult CheckAction(const Action* action);
-  void CheckAssertReturnNanCommand(const Action* action);
   void CheckCommand(const Command* command);
 
   void CheckEvent(const Location* loc, const Event* Event);
@@ -230,7 +224,10 @@ class Validator : public ExprVisitor::Delegate {
 Validator::Validator(Errors* errors,
                      const Script* script,
                      const ValidateOptions& options)
-    : options_(options), errors_(errors), script_(script) {
+    : options_(options),
+      errors_(errors),
+      script_(script),
+      typechecker_(options.features) {
   typechecker_.set_error_callback(
       [this](const char* msg) { OnTypecheckerError(msg); });
 }
@@ -265,8 +262,9 @@ Result Validator::CheckVar(Index max_index,
     }
     return Result::Ok;
   }
-  PrintError(&var->loc, "%s variable out of range (max %" PRIindex ")", desc,
-             max_index);
+  PrintError(&var->loc,
+             "%s variable out of range: %" PRIindex " (max %" PRIindex ")",
+             desc, var->index(), max_index - 1);
   return Result::Error;
 }
 
@@ -446,13 +444,13 @@ void Validator::CheckTypes(const Location* loc,
   }
 }
 
-void Validator::CheckConstTypes(const Location* loc,
-                                const TypeVector& actual,
-                                const ConstVector& expected,
-                                const char* desc) {
+void Validator::CheckResultTypes(const Location* loc,
+                                 const TypeVector& actual,
+                                 const TypeVector& expected,
+                                 const char* desc) {
   if (actual.size() == expected.size()) {
     for (size_t i = 0; i < actual.size(); ++i) {
-      CheckTypeIndex(loc, actual[i], expected[i].type, desc, i, "result");
+      CheckTypeIndex(loc, actual[i], expected[i], desc, i, "result");
     }
   } else {
     PrintError(loc, "expected %" PRIzd " results, got %" PRIzd, expected.size(),
@@ -460,28 +458,6 @@ void Validator::CheckConstTypes(const Location* loc,
   }
 }
 
-void Validator::CheckConstType(const Location* loc,
-                               Type actual,
-                               const ConstVector& expected,
-                               const char* desc) {
-  TypeVector actual_types;
-  if (actual != Type::Void) {
-    actual_types.push_back(actual);
-  }
-  CheckConstTypes(loc, actual_types, expected, desc);
-}
-
-void Validator::CheckAssertReturnNanType(const Location* loc,
-                                         Type actual,
-                                         const char* desc) {
-  // When using assert_return_nan, the result can be either a f32 or f64 type
-  // so we special case it here.
-  if (actual != Type::F32 && actual != Type::F64) {
-    PrintError(loc, "type mismatch at %s. got %s, expected f32 or f64", desc,
-               GetTypeName(actual));
-  }
-}
-
 void Validator::CheckExprList(const Location* loc, const ExprList& exprs) {
   ExprVisitor visitor(this);
   // TODO(binji): Add const-visitors.
@@ -792,37 +768,64 @@ Result Validator::OnElemDropExpr(ElemDropExpr* expr) {
 
 Result Validator::OnTableInitExpr(TableInitExpr* expr) {
   expr_loc_ = &expr->loc;
-  CheckHasTable(&expr->loc, Opcode::TableInit);
-  CheckElemSegmentVar(&expr->var);
-  typechecker_.OnTableInit(expr->var.index());
+  typechecker_.OnTableInit(expr->table_index.index(),
+                           expr->segment_index.index());
+  if (!CheckHasTable(&expr->loc, Opcode::TableInit))
+    return Result::Ok;
+  CheckTableVar(&expr->table_index, nullptr);
+  CheckElemSegmentVar(&expr->segment_index);
   return Result::Ok;
 }
 
 Result Validator::OnTableGetExpr(TableGetExpr* expr) {
+  const Table* table;
+  CheckTableVar(&expr->var, &table);
   expr_loc_ = &expr->loc;
   CheckHasTable(&expr->loc, Opcode::TableGet, expr->var.index());
-  typechecker_.OnTableGet(expr->var.index());
+  typechecker_.OnTableGet(table->elem_type);
   return Result::Ok;
 }
 
 Result Validator::OnTableSetExpr(TableSetExpr* expr) {
+  const Table* table;
+  CheckTableVar(&expr->var, &table);
   expr_loc_ = &expr->loc;
   CheckHasTable(&expr->loc, Opcode::TableSet, expr->var.index());
-  typechecker_.OnTableSet(expr->var.index());
+  typechecker_.OnTableSet(table->elem_type);
   return Result::Ok;
 }
 
 Result Validator::OnTableGrowExpr(TableGrowExpr* expr) {
+  const Table* table;
+  CheckTableVar(&expr->var, &table);
   expr_loc_ = &expr->loc;
   CheckHasTable(&expr->loc, Opcode::TableGrow, expr->var.index());
-  typechecker_.OnTableGrow(expr->var.index());
+  typechecker_.OnTableGrow(table->elem_type);
   return Result::Ok;
 }
 
 Result Validator::OnTableSizeExpr(TableSizeExpr* expr) {
   expr_loc_ = &expr->loc;
   CheckHasTable(&expr->loc, Opcode::TableSize, expr->var.index());
-  typechecker_.OnTableSize(expr->var.index());
+  typechecker_.OnTableSize();
+  return Result::Ok;
+}
+
+Result Validator::OnTableFillExpr(TableFillExpr* expr) {
+  const Table* table;
+  CheckTableVar(&expr->var, &table);
+  expr_loc_ = &expr->loc;
+  CheckHasTable(&expr->loc, Opcode::TableFill, expr->var.index());
+  typechecker_.OnTableFill(table->elem_type);
+  return Result::Ok;
+}
+
+Result Validator::OnRefFuncExpr(RefFuncExpr* expr) {
+  expr_loc_ = &expr->loc;
+  const Func* callee;
+  if (Succeeded(CheckFuncVar(&expr->var, &callee))) {
+    typechecker_.OnRefFuncExpr(expr->var.index());
+  }
   return Result::Ok;
 }
 
@@ -870,8 +873,8 @@ Result Validator::OnReturnCallIndirectExpr(ReturnCallIndirectExpr* expr) {
 
 Result Validator::OnSelectExpr(SelectExpr* expr) {
   expr_loc_ = &expr->loc;
-  typechecker_.OnSelect();
-  return Result::Ok;
+  assert(expr->result_type.size());
+  return typechecker_.OnSelect(expr->result_type[0]);
 }
 
 Result Validator::OnStoreExpr(StoreExpr* expr) {
@@ -1069,10 +1072,14 @@ void Validator::CheckConstInitExpr(const Location* loc,
         break;
       }
 
+      case ExprType::RefFunc:
+        type = Type::Funcref;
+        break;
+
       case ExprType::RefNull:
         type = Type::Nullref;
         break;
-      
+
       default:
         PrintConstExprError(loc, desc);
         return;
@@ -1124,26 +1131,33 @@ void Validator::CheckTable(const Location* loc, const Table* table) {
       !options_.features.reference_types_enabled()) {
     PrintError(loc, "tables must have anyref type");
   }
-  if (table->elem_type != Type::Anyref && table->elem_type != Type::Funcref) {
-    PrintError(loc, "tables must have anyref or funcref type");
+  if (!IsRefType(table->elem_type)) {
+    PrintError(loc, "tables must have reference types");
   }
 }
 
 void Validator::CheckElemSegments(const Module* module) {
+  bool bulk_or_ref_enabled = options_.features.bulk_memory_enabled() ||
+                             options_.features.reference_types_enabled();
+
   for (const ModuleField& field : module->fields) {
     if (auto elem_segment_field = dyn_cast<ElemSegmentModuleField>(&field)) {
       auto&& elem_segment = elem_segment_field->elem_segment;
-      const Table* table;
       for (const ElemExpr& elem_expr : elem_segment.elem_exprs) {
         if (elem_expr.kind == ElemExprKind::RefFunc) {
           CheckFuncVar(&elem_expr.var, nullptr);
+        } else {
+          assert(elem_expr.kind == ElemExprKind::RefNull);
+          if (!bulk_or_ref_enabled) {
+            PrintError(&field.loc, "ref.null is not allowed");
+          }
         }
       }
 
-      if (elem_segment.passive)  {
+      if (elem_segment.kind == SegmentKind::Passive) {
         continue;
       }
-      if (Failed(CheckTableVar(&elem_segment.table_var, &table))) {
+      if (Failed(CheckTableVar(&elem_segment.table_var, nullptr))) {
         continue;
       }
       CheckConstInitExpr(&field.loc, elem_segment.offset, Type::I32,
@@ -1172,7 +1186,7 @@ void Validator::CheckDataSegments(const Module* module) {
     if (auto data_segment_field = dyn_cast<DataSegmentModuleField>(&field)) {
       auto&& data_segment = data_segment_field->data_segment;
       const Memory* memory;
-      if (data_segment.passive)  {
+      if (data_segment.kind == SegmentKind::Passive) {
         continue;
       }
       if (Failed(CheckMemoryVar(&data_segment.memory_var, &memory))) {
@@ -1449,29 +1463,6 @@ Validator::ActionResult Validator::CheckAction(const Action* action) {
   return result;
 }
 
-void Validator::CheckAssertReturnNanCommand(const Action* action) {
-  ActionResult result = CheckAction(action);
-
-  // A valid result type will either be f32 or f64; convert a Types result into
-  // a Type result, so it is easier to check below. Type::Any is used to
-  // specify a type that should not be checked (because an earlier error
-  // occurred).
-  if (result.kind == ActionResult::Kind::Types) {
-    if (result.types->size() == 1) {
-      result.kind = ActionResult::Kind::Type;
-      result.type = (*result.types)[0];
-    } else {
-      PrintError(&action->loc, "expected 1 result, got %" PRIzd,
-                 result.types->size());
-      result.type = Type::Any;
-    }
-  }
-
-  if (result.kind == ActionResult::Kind::Type && result.type != Type::Any) {
-    CheckAssertReturnNanType(&action->loc, result.type, "action");
-  }
-}
-
 void Validator::CheckCommand(const Command* command) {
   switch (command->type) {
     case CommandType::Module:
@@ -1495,15 +1486,19 @@ void Validator::CheckCommand(const Command* command) {
       auto* assert_return_command = cast<AssertReturnCommand>(command);
       const Action* action = assert_return_command->action.get();
       ActionResult result = CheckAction(action);
+      // Here we take the concrete expected output types verify those actains
+      // the types that are the result of the action.
+      TypeVector actual_types;
+      for (auto ex : assert_return_command->expected) {
+        actual_types.push_back(ex.type);
+      }
       switch (result.kind) {
         case ActionResult::Kind::Types:
-          CheckConstTypes(&action->loc, *result.types,
-                          assert_return_command->expected, "action");
+          CheckResultTypes(&action->loc, actual_types, *result.types, "action");
           break;
 
         case ActionResult::Kind::Type:
-          CheckConstType(&action->loc, result.type,
-                         assert_return_command->expected, "action");
+          CheckResultTypes(&action->loc, actual_types, {result.type}, "action");
           break;
 
         case ActionResult::Kind::Error:
@@ -1513,16 +1508,6 @@ void Validator::CheckCommand(const Command* command) {
       break;
     }
 
-    case CommandType::AssertReturnCanonicalNan:
-      CheckAssertReturnNanCommand(
-          cast<AssertReturnCanonicalNanCommand>(command)->action.get());
-      break;
-
-    case CommandType::AssertReturnArithmeticNan:
-      CheckAssertReturnNanCommand(
-          cast<AssertReturnArithmeticNanCommand>(command)->action.get());
-      break;
-
     case CommandType::AssertTrap:
       // ignore result type.
       CheckAction(cast<AssertTrapCommand>(command)->action.get());
