@@ -21,6 +21,7 @@
 #include "src/cast.h"
 #include "src/expr-visitor.h"
 #include "src/make-unique.h"
+#include "src/stream.h"
 #include "src/utf8.h"
 
 #define WABT_TRACING 0
@@ -143,8 +144,10 @@ bool IsPlainInstr(TokenType token_type) {
     case TokenType::TableSet:
     case TokenType::TableGrow:
     case TokenType::TableSize:
+    case TokenType::TableFill:
     case TokenType::Throw:
     case TokenType::Rethrow:
+    case TokenType::RefFunc:
     case TokenType::RefNull:
     case TokenType::RefIsNull:
     case TokenType::AtomicLoad:
@@ -219,8 +222,6 @@ bool IsCommand(TokenTypePair pair) {
     case TokenType::AssertInvalid:
     case TokenType::AssertMalformed:
     case TokenType::AssertReturn:
-    case TokenType::AssertReturnArithmeticNan:
-    case TokenType::AssertReturnCanonicalNan:
     case TokenType::AssertTrap:
     case TokenType::AssertUnlinkable:
     case TokenType::Get:
@@ -540,12 +541,14 @@ Result WastParser::ErrorIfLpar(const std::vector<std::string>& expected,
   return Result::Ok;
 }
 
-void WastParser::ParseBindVarOpt(std::string* name) {
+bool WastParser::ParseBindVarOpt(std::string* name) {
   WABT_TRACE(ParseBindVarOpt);
-  if (PeekMatch(TokenType::Var)) {
-    Token token = Consume();
-    *name = token.text().to_string();
+  if (!PeekMatch(TokenType::Var)) {
+    return false;
   }
+  Token token = Consume();
+  *name = token.text().to_string();
+  return true;
 }
 
 Result WastParser::ParseVar(Var* out_var) {
@@ -611,6 +614,7 @@ bool WastParser::ParseOffsetExprOpt(ExprList* out_expr_list) {
 Result WastParser::ParseTextList(std::vector<uint8_t>* out_data) {
   WABT_TRACE(ParseTextList);
   if (!ParseTextListOpt(out_data)) {
+    // TODO(binji): Add error message here.
     return Result::Error;
   }
 
@@ -633,15 +637,40 @@ Result WastParser::ParseVarList(VarVector* out_var_list) {
   while (ParseVarOpt(&var)) {
     out_var_list->emplace_back(var);
   }
-  return out_var_list->empty() ? Result::Error : Result::Ok;
+  if (out_var_list->empty()) {
+    return ErrorExpected({"a var"}, "12 or $foo");
+  } else {
+    return Result::Ok;
+  }
 }
 
-Result WastParser::ParseElemExprVarList(ElemExprVector* out_list) {
-  WABT_TRACE(ParseElemExprVarList);
-  if (!ParseElemExprVarListOpt(out_list)) {
-    return Result::Error;
+bool WastParser::ParseElemExprOpt(ElemExpr* out_elem_expr) {
+  Location loc = GetLocation();
+  if (MatchLpar(TokenType::RefNull)) {
+    if (!(options_->features.bulk_memory_enabled() ||
+          options_->features.reference_types_enabled())) {
+      Error(loc, "ref.null not allowed");
+    }
+    *out_elem_expr = ElemExpr();
+    EXPECT(Rpar);
+    return true;
+  } else if (MatchLpar(TokenType::RefFunc)) {
+    Var var;
+    CHECK_RESULT(ParseVar(&var));
+    *out_elem_expr = ElemExpr(var);
+    EXPECT(Rpar);
+    return true;
+  } else {
+    return false;
   }
-  return Result::Ok;
+}
+
+bool WastParser::ParseElemExprListOpt(ElemExprVector* out_list) {
+  ElemExpr elem_expr;
+  while (ParseElemExprOpt(&elem_expr)) {
+    out_list->push_back(elem_expr);
+  }
+  return !out_list->empty();
 }
 
 bool WastParser::ParseElemExprVarListOpt(ElemExprVector* out_list) {
@@ -667,6 +696,9 @@ Result WastParser::ParseValueType(Type* out_type) {
       is_enabled = options_->features.simd_enabled();
       break;
     case Type::Anyref:
+    case Type::Funcref:
+    case Type::Hostref:
+    case Type::Exnref:
       is_enabled = options_->features.reference_types_enabled();
       break;
     default:
@@ -694,7 +726,7 @@ Result WastParser::ParseValueTypeList(TypeVector* out_type_list) {
 Result WastParser::ParseRefType(Type* out_type) {
   WABT_TRACE(ParseRefType);
   if (!PeekMatch(TokenType::ValueType)) {
-    return ErrorExpected({"anyref", "funcref"});
+    return ErrorExpected({"funcref", "anyref", "nullref", "exnref"});
   }
 
   Token token = Consume();
@@ -915,10 +947,21 @@ Result WastParser::ParseDataModuleField(Module* module) {
   ParseBindVarOpt(&name);
   auto field = MakeUnique<DataSegmentModuleField>(loc, name);
 
-  if (ParseVarOpt(&field->data_segment.memory_var, Var(0, loc))) {
+  if (PeekMatchLpar(TokenType::Memory)) {
+    EXPECT(Lpar);
+    EXPECT(Memory);
+    CHECK_RESULT(ParseVar(&field->data_segment.memory_var));
+    EXPECT(Rpar);
+    CHECK_RESULT(ParseOffsetExpr(&field->data_segment.offset));
+  } else if (ParseVarOpt(&field->data_segment.memory_var, Var(0, loc))) {
     CHECK_RESULT(ParseOffsetExpr(&field->data_segment.offset));
   } else if (!ParseOffsetExprOpt(&field->data_segment.offset)) {
-    field->data_segment.passive = true;
+    if (!options_->features.bulk_memory_enabled()) {
+      Error(loc, "passive data segments are not allowed");
+      return Result::Error;
+    }
+
+    field->data_segment.kind = SegmentKind::Passive;
   }
 
   ParseTextListOpt(&field->data_segment.data);
@@ -932,35 +975,50 @@ Result WastParser::ParseElemModuleField(Module* module) {
   EXPECT(Lpar);
   Location loc = GetLocation();
   EXPECT(Elem);
-  std::string name;
-  ParseBindVarOpt(&name);
-  auto field = MakeUnique<ElemSegmentModuleField>(loc, name);
 
-  if (ParseRefTypeOpt(&field->elem_segment.elem_type)) {
-    field->elem_segment.passive = true;
-    // Parse a potentially empty sequence of ElemExprs.
-    while (true) {
-      Var var;
-      if (MatchLpar(TokenType::RefNull)) {
-        field->elem_segment.elem_exprs.emplace_back();
-        EXPECT(Rpar);
-      } else if (MatchLpar(TokenType::RefFunc)) {
-        CHECK_RESULT(ParseVar(&var));
-        field->elem_segment.elem_exprs.emplace_back(var);
-        EXPECT(Rpar);
-      } else if (ParseVarOpt(&var)) {
-        // TODO: This format will be removed by
-        // https://github.com/WebAssembly/bulk-memory-operations/pull/84
-        field->elem_segment.elem_exprs.emplace_back(var);
-      } else {
-        CHECK_RESULT(ErrorIfLpar({"ref.null", "ref.func"}));
-        break;
-      }
+  // With MVP text format the name here was intended to refer to the table
+  // that the elem segment was part of, but we never did anything with this name
+  // since there was only one table anyway.
+  // With bulk-memory enabled this introduces a new name for the particular
+  // elem segment.
+  std::string initial_name;
+  bool has_name = ParseBindVarOpt(&initial_name);
+
+  std::string segment_name = initial_name;
+  if (!options_->features.bulk_memory_enabled()) {
+    segment_name = "";
+  }
+  auto field = MakeUnique<ElemSegmentModuleField>(loc, segment_name);
+
+  // Optional table specifier
+  if (options_->features.bulk_memory_enabled()) {
+    if (PeekMatchLpar(TokenType::Table)) {
+      EXPECT(Lpar);
+      EXPECT(Table);
+      CHECK_RESULT(ParseVar(&field->elem_segment.table_var));
+      EXPECT(Rpar);
+    } else {
+      ParseVarOpt(&field->elem_segment.table_var, Var(0, loc));
     }
+  } else {
+    if (has_name) {
+      field->elem_segment.table_var = Var(initial_name, loc);
+    } else {
+      ParseVarOpt(&field->elem_segment.table_var, Var(0, loc));
+    }
+  }
+
+  if (!ParseOffsetExprOpt(&field->elem_segment.offset)) {
+    field->elem_segment.kind = SegmentKind::Passive;
+  }
+
+  if (ParseRefTypeOpt(&field->elem_segment.elem_type)) {
+    ParseElemExprListOpt(&field->elem_segment.elem_exprs);
   } else {
     field->elem_segment.elem_type = Type::Funcref;
-    ParseVarOpt(&field->elem_segment.table_var, Var(0, loc));
-    CHECK_RESULT(ParseOffsetExpr(&field->elem_segment.offset));
+    if (PeekMatch(TokenType::Func)) {
+      EXPECT(Func);
+    }
     ParseElemExprVarListOpt(&field->elem_segment.elem_exprs);
   }
   EXPECT(Rpar);
@@ -1224,6 +1282,10 @@ Result WastParser::ParseStartModuleField(Module* module) {
   WABT_TRACE(ParseStartModuleField);
   EXPECT(Lpar);
   Location loc = GetLocation();
+  if (module->starts.size() > 0) {
+    Error(loc, "multiple start sections");
+    return Result::Error;
+  }
   EXPECT(Start);
   Var var;
   CHECK_RESULT(ParseVar(&var));
@@ -1264,7 +1326,17 @@ Result WastParser::ParseTableModuleField(Module* module) {
     elem_segment.table_var = Var(module->tables.size());
     elem_segment.offset.push_back(MakeUnique<ConstExpr>(Const::I32(0)));
     elem_segment.offset.back().loc = loc;
-    CHECK_RESULT(ParseElemExprVarList(&elem_segment.elem_exprs));
+    elem_segment.elem_type = elem_type;
+    // Syntax is either an optional list of var (legacy), or a non-empty list
+    // of elem expr.
+    ElemExpr elem_expr;
+    if (ParseElemExprOpt(&elem_expr)) {
+      elem_segment.elem_exprs.push_back(elem_expr);
+      // Parse the rest.
+      ParseElemExprListOpt(&elem_segment.elem_exprs);
+    } else {
+      ParseElemExprVarListOpt(&elem_segment.elem_exprs);
+    }
     EXPECT(Rpar);
 
     auto table_field = MakeUnique<TableModuleField>(loc, name);
@@ -1480,10 +1552,19 @@ Result WastParser::ParsePlainInstr(std::unique_ptr<Expr>* out_expr) {
       out_expr->reset(new DropExpr(loc));
       break;
 
-    case TokenType::Select:
+    case TokenType::Select: {
       Consume();
-      out_expr->reset(new SelectExpr(loc));
+      TypeVector result;
+      if (options_->features.reference_types_enabled() &&
+          MatchLpar(TokenType::Result)) {
+        CHECK_RESULT(ParseValueTypeList(&result));
+        EXPECT(Rpar);
+      } else {
+        result.push_back(Type::Any);
+      }
+      out_expr->reset(new SelectExpr(result, loc));
       break;
+    }
 
     case TokenType::Br:
       Consume();
@@ -1527,9 +1608,9 @@ Result WastParser::ParsePlainInstr(std::unique_ptr<Expr>* out_expr) {
     case TokenType::CallIndirect: {
       Consume();
       auto expr = MakeUnique<CallIndirectExpr>(loc);
+      ParseVarOpt(&expr->table, Var(0));
       CHECK_RESULT(ParseTypeUseOpt(&expr->decl));
       CHECK_RESULT(ParseUnboundFuncSignature(&expr->decl.sig));
-      ParseVarOpt(&expr->table, Var(0));
       *out_expr = std::move(expr);
       break;
     }
@@ -1586,7 +1667,7 @@ Result WastParser::ParsePlainInstr(std::unique_ptr<Expr>* out_expr) {
 
     case TokenType::Const: {
       Const const_;
-      CHECK_RESULT(ParseConst(&const_));
+      CHECK_RESULT(ParseConst(&const_, ConstType::Normal));
       out_expr->reset(new ConstExpr(const_, loc));
       break;
     }
@@ -1646,20 +1727,34 @@ Result WastParser::ParsePlainInstr(std::unique_ptr<Expr>* out_expr) {
       out_expr->reset(new MemoryGrowExpr(loc));
       break;
 
-    case TokenType::TableCopy:
+    case TokenType::TableCopy: {
       ErrorUnlessOpcodeEnabled(Consume());
-      out_expr->reset(new TableCopyExpr(loc));
+      Var dst(0);
+      Var src(0);
+      if (options_->features.reference_types_enabled()) {
+      // TODO: disabled for now, since the spec tests don't currently use.
+#if 0
+        CHECK_RESULT(ParseVar(&dst));
+        CHECK_RESULT(ParseVar(&src));
+#endif
+      }
+      out_expr->reset(new TableCopyExpr(dst, src, loc));
       break;
+    }
 
     case TokenType::ElemDrop:
       ErrorUnlessOpcodeEnabled(Consume());
       CHECK_RESULT(ParsePlainInstrVar<ElemDropExpr>(loc, out_expr));
       break;
 
-    case TokenType::TableInit:
+    case TokenType::TableInit: {
       ErrorUnlessOpcodeEnabled(Consume());
-      CHECK_RESULT(ParsePlainInstrVar<TableInitExpr>(loc, out_expr));
+      Var segment_index(0);
+      CHECK_RESULT(ParseVar(&segment_index));
+      Var table_index(0);
+      out_expr->reset(new TableInitExpr(segment_index, table_index, loc));
       break;
+    }
 
     case TokenType::TableGet:
       ErrorUnlessOpcodeEnabled(Consume());
@@ -1681,6 +1776,16 @@ Result WastParser::ParsePlainInstr(std::unique_ptr<Expr>* out_expr) {
       CHECK_RESULT(ParsePlainInstrVar<TableSizeExpr>(loc, out_expr));
       break;
 
+    case TokenType::TableFill:
+      ErrorUnlessOpcodeEnabled(Consume());
+      CHECK_RESULT(ParsePlainInstrVar<TableFillExpr>(loc, out_expr));
+      break;
+
+    case TokenType::RefFunc:
+      ErrorUnlessOpcodeEnabled(Consume());
+      CHECK_RESULT(ParsePlainInstrVar<RefFuncExpr>(loc, out_expr));
+      break;
+
     case TokenType::RefNull:
       ErrorUnlessOpcodeEnabled(Consume());
       out_expr->reset(new RefNullExpr(loc));
@@ -1910,12 +2015,29 @@ Result WastParser::ParseSimdV128Const(Const* const_, TokenType token_type) {
     }
   }
 
-  memcpy(&const_->v128_bits.v, v128_bytes.data(), 16);
+  memcpy(&const_->vec128.v, v128_bytes.data(), 16);
 
   return Result::Ok;
 }
 
-Result WastParser::ParseConst(Const* const_) {
+Result WastParser::ParseExpectedNan(ExpectedNan* expected) {
+  WABT_TRACE(ParseExpectedNan);
+  TokenType token_type = Peek();
+  switch (token_type) {
+    case TokenType::NanArithmetic:
+      *expected = ExpectedNan::Arithmetic;
+      break;
+    case TokenType::NanCanonical:
+      *expected = ExpectedNan::Canonical;
+      break;
+    default:
+      return Result::Error;
+  }
+  Consume();
+  return Result::Ok;
+}
+
+Result WastParser::ParseConst(Const* const_, ConstType type) {
   WABT_TRACE(ParseConst);
   Token token = Consume();
   Opcode opcode = token.opcode();
@@ -1938,6 +2060,9 @@ Result WastParser::ParseConst(Const* const_) {
       end = sv.end();
       break;
     }
+    case TokenType::NanArithmetic:
+    case TokenType::NanCanonical:
+      break;
     default:
       return ErrorExpected({"a numeric literal"}, "123, -45, 6.7e8");
     }
@@ -1959,11 +2084,21 @@ Result WastParser::ParseConst(Const* const_) {
 
     case Opcode::F32Const:
       const_->type = Type::F32;
+      if (type == ConstType::Expectation &&
+          ParseExpectedNan(&const_->expected) == Result::Ok) {
+        const_->is_expected_nan = true;
+        break;
+      }
       result = ParseFloat(literal.type, s, end, &const_->f32_bits);
       break;
 
     case Opcode::F64Const:
       const_->type = Type::F64;
+      if (type == ConstType::Expectation &&
+          ParseExpectedNan(&const_->expected) == Result::Ok) {
+        const_->is_expected_nan = true;
+        break;
+      }
       result = ParseDouble(literal.type, s, end, &const_->f64_bits);
       break;
 
@@ -2025,7 +2160,7 @@ Result WastParser::ParseHostRef(Const* const_) {
   uint64_t ref_bits;
   Result result = ParseInt64(s, end, &ref_bits, ParseIntType::UnsignedOnly);
 
-  const_->type = ref_bits ? Type::Anyref : Type::Nullref;
+  const_->type = Type::Hostref;
   const_->ref_bits = static_cast<uintptr_t>(ref_bits);
 
   if (Failed(result)) {
@@ -2038,15 +2173,16 @@ Result WastParser::ParseHostRef(Const* const_) {
   return Result::Ok;
 }
 
-Result WastParser::ParseConstList(ConstVector* consts) {
+Result WastParser::ParseConstList(ConstVector* consts, ConstType type) {
   WABT_TRACE(ParseConstList);
   while (PeekMatchLpar(TokenType::Const) || PeekMatchLpar(TokenType::RefNull) ||
-         PeekMatchLpar(TokenType::RefHost)) {
+         PeekMatchLpar(TokenType::RefHost) ||
+         PeekMatchLpar(TokenType::RefFunc)) {
     Consume();
     Const const_;
     switch (Peek()) {
       case TokenType::Const:
-        CHECK_RESULT(ParseConst(&const_));
+        CHECK_RESULT(ParseConst(&const_, type));
         break;
       case TokenType::RefNull: {
         auto token = Consume();
@@ -2056,6 +2192,14 @@ Result WastParser::ParseConstList(ConstVector* consts) {
         const_.ref_bits = 0;
         break;
       }
+      case TokenType::RefFunc: {
+        auto token = Consume();
+        ErrorUnlessOpcodeEnabled(token);
+        const_.loc = GetLocation();
+        const_.type = Type::Funcref;
+        const_.ref_bits = 0;
+        break;
+      }
       case TokenType::RefHost:
         CHECK_RESULT(ParseHostRef(&const_));
         break;
@@ -2343,12 +2487,6 @@ Result WastParser::ParseCommand(Script* script, CommandPtr* out_command) {
     case TokenType::AssertReturn:
       return ParseAssertReturnCommand(out_command);
 
-    case TokenType::AssertReturnArithmeticNan:
-      return ParseAssertReturnArithmeticNanCommand(out_command);
-
-    case TokenType::AssertReturnCanonicalNan:
-      return ParseAssertReturnCanonicalNanCommand(out_command);
-
     case TokenType::AssertTrap:
       return ParseAssertTrapCommand(out_command);
 
@@ -2395,26 +2533,12 @@ Result WastParser::ParseAssertReturnCommand(CommandPtr* out_command) {
   EXPECT(AssertReturn);
   auto command = MakeUnique<AssertReturnCommand>();
   CHECK_RESULT(ParseAction(&command->action));
-  CHECK_RESULT(ParseConstList(&command->expected));
+  CHECK_RESULT(ParseConstList(&command->expected, ConstType::Expectation));
   EXPECT(Rpar);
   *out_command = std::move(command);
   return Result::Ok;
 }
 
-Result WastParser::ParseAssertReturnArithmeticNanCommand(
-    CommandPtr* out_command) {
-  WABT_TRACE(ParseAssertReturnArithmeticNanCommand);
-  return ParseAssertActionCommand<AssertReturnArithmeticNanCommand>(
-      TokenType::AssertReturnArithmeticNan, out_command);
-}
-
-Result WastParser::ParseAssertReturnCanonicalNanCommand(
-    CommandPtr* out_command) {
-  WABT_TRACE(ParseAssertReturnCanonicalNanCommand);
-  return ParseAssertActionCommand<AssertReturnCanonicalNanCommand>(
-      TokenType::AssertReturnCanonicalNan, out_command);
-}
-
 Result WastParser::ParseAssertTrapCommand(CommandPtr* out_command) {
   WABT_TRACE(ParseAssertTrapCommand);
   EXPECT(Lpar);
@@ -2464,6 +2588,10 @@ Result WastParser::ParseModuleCommand(Script* script, CommandPtr* out_command) {
     case ScriptModuleType::Binary: {
       auto* bsm = cast<BinaryScriptModule>(script_module.get());
       ReadBinaryOptions options;
+#if WABT_TRACING
+      auto log_stream = FileStream::CreateStdout();
+      options.log_stream = log_stream.get();
+#endif
       options.features = options_->features;
       Errors errors;
       const char* filename = "<text>";
@@ -2528,7 +2656,7 @@ Result WastParser::ParseAction(ActionPtr* out_action) {
       auto action = MakeUnique<InvokeAction>(loc);
       ParseVarOpt(&action->module_var, Var(last_module_index_, loc));
       CHECK_RESULT(ParseQuotedText(&action->name));
-      CHECK_RESULT(ParseConstList(&action->args));
+      CHECK_RESULT(ParseConstList(&action->args, ConstType::Normal));
       *out_action = std::move(action);
       break;
     }
@@ -2562,6 +2690,8 @@ Result WastParser::ParseScriptModule(
     case TokenType::Bin: {
       Consume();
       std::vector<uint8_t> data;
+      // TODO(binji): The spec allows this to be empty, switch to
+      // ParseTextListOpt.
       CHECK_RESULT(ParseTextList(&data));
 
       auto bsm = MakeUnique<BinaryScriptModule>();
@@ -2575,6 +2705,8 @@ Result WastParser::ParseScriptModule(
     case TokenType::Quote: {
       Consume();
       std::vector<uint8_t> data;
+      // TODO(binji): The spec allows this to be empty, switch to
+      // ParseTextListOpt.
       CHECK_RESULT(ParseTextList(&data));
 
       auto qsm = MakeUnique<QuotedScriptModule>();
